
import os
import sys
import datetime
from typing import Optional
import re

from peewee import DatabaseError
from OpenSSL import crypto
from OpenSSL.SSL import FILETYPE_PEM

from .models import CA, Config, ConfigType, Certificate, CRL, database, DATABASE_FILENAME
from .utils import ask_pem_format, ask_private_key_passphrase, get_csr_san, generate_random_serial_number, \
     get_cert_text, get_csr_text, define_san, PEMFormatType, validate_or_build_san, regex_domain, confirm, ask, \
     ask_password, ask_date_in_future, nb_seconds_to_date, print_tabulated, \
     get_csr_cn, write_cert_pem, write_private_key_pem, write_p12, ask_cn_and_san

"""
Valid key sizes
"""
VALID_KEY_SIZES = [1024, 2048, 4096, 8192]

"""
Valid hash algorithms
"""
VALID_HASH_ALGOS = ['sha1', 'sha256', 'sha512']


class CAManager:
    """
    Certificate Authority Manager
    """
    def __init__(self):
        """
        Constructor.

        By default, the manager is not loaded. This is useful for the first start because the vault must be created.
        """
        self.ca_name = 'Default'

        # Containing the Certificate Authority when loaded from the vault
        self.ca = None

    def create_vault(self, vault_password: Optional[str] = None) -> None:
        """
        Setup the vault for the first use.

        The vault cannot be existing. The user is prompted to enter :
            - a password to encrypt the vault
            - the CA certificate
            - the CA key (and the password protecting the key if any)
        """
        if os.path.isfile(DATABASE_FILENAME):
            sys.stderr.write(f'Vault "{DATABASE_FILENAME}" is already existing\n')
            exit(-1)

        vault_password = ask_password('Enter the password that will be used to encrypt the CA vault')
        vault_password_confirm = ask_password('Confirm it')

        if vault_password != vault_password_confirm:
            sys.stderr.write(f'The two provided passwords are not matching\n')
            exit(-1)

        database.init(DATABASE_FILENAME, passphrase=vault_password)

        with database:
            # Create table
            database.create_tables([CA, Certificate, CRL, Config])

            # Read CA pem from stdin
            try:
                pem_ca_cert = ask_pem_format('Paste your CA certificate in PEM format', PEMFormatType.CERT)
                ca_cert = crypto.load_certificate(FILETYPE_PEM, pem_ca_cert.encode('utf8'))
            except ValueError:
                sys.stderr.write(f'Invalid PEM certificate')
                exit(-1)
            except crypto.Error as e:
                sys.stderr.write(f'Invalid PEM certificate. OpenSSL returns with an error.')
                exit(-1)

            # Read CA key from stdin
            try:
                pem_ca_key = ask_pem_format('Paste your CA key in PEM format', PEMFormatType.KEY)
                ca_key = crypto.load_privatekey(FILETYPE_PEM, pem_ca_key.encode('utf8'),
                                                passphrase=ask_private_key_passphrase)
            except ValueError:
                sys.stderr.write(f'Invalid PEM key')
                exit(-1)
            except crypto.Error as e:
                sys.stderr.write(f'Invalid PEM key. OpenSSL returns with an error.')
                exit(-1)

            # Check CA cert and key correspondence
            if crypto.dump_publickey(FILETYPE_PEM, ca_cert.get_pubkey()) != crypto.dump_publickey(FILETYPE_PEM, ca_key):
                sys.stderr.write(f'The certificate and the private key are not corresponding')
                exit(-1)

            # Create the CA in the database
            CA.create(
                name=self.ca_name,
                cert=pem_ca_cert,
                key=pem_ca_key,
            )

            # Create the config
            Config.create(
                name='default_key_size',
                type=ConfigType.INT,
                value=4096
            )

            Config.create(
                name='default_hash_algo',
                type=ConfigType.STRING,
                value='sha256'
            )

            Config.create(
                name='default_validity_seconds',
                type=ConfigType.INT,
                value=3 * 365 * 24 * 60 * 60  # 3 years
            )

        print('The vault has been successfully created.')

    def load(self):
        """
        Load the vault.

        The password is asking.
        """
        if not os.path.isfile(DATABASE_FILENAME):
            sys.stderr.write('The vault doesn\'t exist yer. Please run with --setup to initialize it.\n')
            exit(-1)

        # Ask the vault password and initialize the database with it
        database.init(DATABASE_FILENAME, passphrase=ask_password('Enter the CA vault password'))

        try:
            database.get_tables()
        except DatabaseError as exc:
            if exc.args[0] == 'file is not a database':
                database.init(None)

                sys.stderr.write('The provided password is incorrect\n')
                exit(-1)
            else:
                raise exc

    @database.connection_context()
    def list(self, all_certificates: bool = False, only_soon_expired: bool = False):
        """
        Print the list of managed certificates.

        :param all_certificates: also the revoked/expired/renewed certificates
        :param only_soon_expired: only the soon expired certificates
        """
        if all_certificates and only_soon_expired:
            raise RuntimeError('Cannot use all and only_soon_expired')

        certs = Certificate.select()

        if only_soon_expired:
            now = datetime.datetime.now()
            next_month = now + datetime.timedelta(days=31)

            certs = certs.where((Certificate.is_renewed == False) & (Certificate.not_after >= now) &
                                (Certificate.not_after <= next_month) & (Certificate.is_revoked == False))
        elif not all_certificates:
            certs = certs.where((Certificate.is_revoked == False) & (Certificate.is_renewed == False) &
                                (Certificate.not_after > datetime.datetime.now()))

        nb_certs_found = certs.count()

        if nb_certs_found == 0:
            print('No certificate found')
        else:
            print(f'{nb_certs_found} {"certificate" if nb_certs_found < 2 else "certificates"} found :')

            for c in certs.iterator():
                print(f'\t{c}')

    def generate_new_cert(self, cn: Optional[str] = None, san: Optional[str] = None, keysize: Optional[int] = None,
                          hashalgo: Optional[str] = None, expire_date: Optional[str] = None) -> None:
        """
        Generate a new certificate.

        The method accepts parameter that are only be processed if the "cn" is defined. If none of the parameters are
        supplied, the method is interactive and asks the user for the information.

        :param cn: the Common Name
        :param san: Subject Alternative Name
        :param keysize: the keysize
        :param hashalgo: the hash algorithm
        :param expire_date: the expire date
        """

        if not cn:  # The CN is not defined, we're going into interactive mode
            if not confirm(f'Use default params ({self._get_config("default_key_size")} bits - '
                           f'{self._get_config("default_hash_algo")})'):
                keysize = int(ask('Key size', values=VALID_KEY_SIZES))
                hashalgo = ask('Hash algorithm', values=VALID_HASH_ALGOS)

            cn, san = ask_cn_and_san()

            nb_days = self._get_config("default_validity_seconds")
            not_after = datetime.date.today() + datetime.timedelta(seconds=nb_days)
            if not confirm(f'Use the default validity (will expire on {not_after.strftime("%d/%m/%Y")}'):
                expire_date = ask_date_in_future()

        try:
            cert, certificate = self._generate_signed_cert(cn, san, keysize, hashalgo, expire_date, ask_confirm=True)

            print('Certificate successfully created !\n')

            print(get_cert_text(cert), end='')
            print(certificate.cert)
            print(certificate.key)
        except (ValueError, RuntimeError) as e:
            sys.stderr.write(f'Error: {e}.\n')
            sys.stderr.write(f'The certificate cannot be created.\n')
            exit(-1)

    def sign_csr(self, csr_filepath: Optional[str] = None, cn: Optional[str] = None, san: Optional[str] = None,
                 expire_date: Optional[str] = None) -> None:
        """
        Sign a certificate.

        If the filepath is not provided, the content is read from stdin.

        The method accepts parameter that will override the CSR content.

        :param csr_filepath: the Certificate Signing Request filepath
        :param cn: the Common Name
        :param san: Subject Alternative Name
        :param expire_date: the expire date
        """

        if csr_filepath:
            if not os.path.exists(csr_filepath):
                sys.stderr.write(f'The file "{csr_filepath}" doesn\'t exist\n')
                exit(-1)

            with open(csr_filepath, 'rb') as f:
                csr_content = f.read()
        else:
            csr_content = ask_pem_format('Paste your CSR in PEM format', PEMFormatType.CSR).encode('utf8')

        csr = None
        try:
            csr = crypto.load_certificate_request(FILETYPE_PEM, csr_content)
        except ValueError:
            sys.stderr.write(f'Invalid PEM CSR')
            exit(-1)
        except crypto.Error as e:
            sys.stderr.write(f'Invalid PEM CSR. OpenSSL returns with an error.')
            exit(-1)

        print('You\'re going to process the following Certificate Signing Request :\n')
        print_tabulated(get_csr_text(csr))

        default_cn = get_csr_cn(csr)
        default_san = get_csr_san(csr)

        print('\nThis CSR apply to :')
        print(f'\tCommon Name : {default_cn}')
        print(f'\tSubject Alternative Name : {default_san if default_san else "*** not defined ***"}\n')

        print('Please note that the CN will automatically be added to the SAN if it is missing.\n')

        if not cn and not san:
            if confirm(f'Do you want to overwrite theses values'):
                cn, san = ask_cn_and_san()

        if cn and cn != default_cn or san and san != default_san:
            print('\nYou will overwrite the following values :')

            if cn and cn != default_cn:
                print(f'\tCommon Name : {cn}')

            if san and san != default_san:
                print(f'\tSubject Alternative Name : {san}')

        try:
            print('')
            cert, certificate = self._sign_csr(csr, overwrite_cn=cn, overwrite_san=san, ask_confirm=True)

            print('Certificate successfully signed !\n')

            print(get_cert_text(cert), end='')
            print(certificate.cert)

        except (ValueError, RuntimeError) as e:
            sys.stderr.write(f'Error: {e}.\n')
            sys.stderr.write(f'The certificate cannot be signed.')

    @database.connection_context()
    def renew(self, target_certificate: str):
        """
        Renew a certificate.

        :param target_certificate: CN or ID of the certificate
        """
        certificate = self.select_certificate(target_certificate)
        certificate.load_cert()
        pub_key = certificate.cert_obj.get_pubkey()

        # Generate CSR
        req = crypto.X509Req()
        req.get_subject().CN = certificate.cn
        req.set_pubkey(pub_key)

        # Define the SAN
        define_san(req, certificate.san)

        try:
            # Sign the CSR
            signed_cert, vault_certificate = self._sign_csr(req, ask_confirm=True, disable_cert_existence_check=True)

            # Set the previous certificate as renewed
            certificate.is_renewed = True
            certificate.save()

            # Save the private key for the new certificate
            vault_certificate.key = certificate.key
            vault_certificate.save()

            print('Certificate renewed successfully !\n')

            vault_certificate.load_cert()
            print(get_cert_text(vault_certificate.cert_obj), end='')
            print(vault_certificate.cert)

        except (ValueError, RuntimeError) as e:
            sys.stderr.write(f'Error: {e}.\n')
            sys.stderr.write(f'The certificate cannot be renewed.\n')
            exit(-1)

    def export(self, target_certificate: str, output_format: str, out_path: Optional[str]):
        """
        Export the certificate and its private key if available.

        If the output_format is "pem", the data is exported to a file if the out_path is provided (the file extension
        .pem is added for the certificate and .key for the key). If no out_path is provided, the data is printed to
        stdout.

        If the output_format is "p12", the data is exported to the out_path. If the out_path doesn't have the ".p12"
        extension, this method add the extension. The p12_passphrase must be provided.

        :param target_certificate: the certificate CN or ID
        :param output_format: the output format : "pem" or "p12"
        :param out_path: the output path
        """
        if output_format not in ('pem', 'p12'):
            raise ValueError(f'Output format "{output_format}" not supported')

        certificate = self.select_certificate(target_certificate)

        print(f'You\'re going to export the certificate {certificate}...')

        p12_passphrase = None
        if output_format == 'p12':
            p12_passphrase = ask_password('Enter the password that will be used protect the .p12')
            p12_passphrase_confirm = ask_password('Confirm it')

            if p12_passphrase != p12_passphrase_confirm:
                sys.stderr.write(f'The two provided passwords are not matching\n')
                exit(-1)

        try:
            self._export(certificate, output_format, out_path, p12_passphrase)
        except RuntimeError as e:
            sys.stderr.write(f'Error: {str(e)}\n')
            exit(-1)

    def _export(self, certificate: Certificate, output_format: str, out_path: Optional[str],
                p12_passphrase: Optional[str] = None):
        """
        Export the certificate and its private key if available.

        If the output_format is "pem", the data is exported to a file if the out_path is provided (the file extension
        .pem is added for the certificate and .key for the key). If no out_path is provided, the data is printed to
        stdout.

        If the output_format is "p12", the data is exported to the out_path. If the out_path doesn't have the ".p12"
        extension, this method add the extension. The p12_passphrase must be provided.

        :param certificate: the certificate
        :param output_format: the output format : "pem" or "p12"
        :param out_path: the output path
        :param p12_passphrase: the P12 passphrase
        """
        output_format = output_format.lower()
        if output_format not in ('pem', 'p12'):
            raise ValueError(f'Output format "{output_format}" not supported')

        if output_format == 'p12' and not out_path:
            raise ValueError('For the p12 format, the out_path must be specified')

        if output_format == 'p12' and not p12_passphrase:
            raise ValueError('For the p12 format, the p12_passphrase must be specified')

        certificate.load_cert()

        if certificate.key:
            certificate.load_key()

        if output_format == 'pem':
            if out_path:
                out_path_cert = out_path + '.pem'
                out_path_key = out_path + '.key'

                if os.path.exists(out_path_cert):
                    raise RuntimeError(f'The output certificate file "{out_path_cert}" already exists')

                if certificate.key:
                    if os.path.exists(out_path_key):
                        raise RuntimeError(f'The output private key file "{out_path_key}" already exists')

                write_cert_pem(certificate.cert_obj, out_path_cert)

                if certificate.key:
                    write_private_key_pem(certificate.key_obj, out_path_key)

                if certificate.key:
                    print(f'The certificate and its private key have been exported to {out_path}[.pem|.key]')
                else:
                    print(f'The certificate has been exported to {out_path}.pem. There is no private key linked.')
            else:
                print(certificate.cert)

                if certificate.key:
                    print(certificate.key)
                else:
                    print('There is no private key linked.')
        else:
            if not out_path.endswith('.p12'):
                out_path += '.p12'

            write_p12(out_path, p12_passphrase, certificate.cert_obj, certificate.key_obj)

            if certificate.key:
                print(f'The certificate and its private key have been exported to {out_path}')
            else:
                print(f'The certificate has been exported to {out_path}. There is no private key linked.')

        certificate.clear_key()

    def _load_ca(self):
        """
        Load the Certificate Authority from the vault
        """
        if self.ca:
            return

        self.ca = CA.get(CA.name == self.ca_name)
        if not self.ca:
            raise RuntimeError(f'Cannot find the CA with the name "{self.ca_name}"')

        self.ca.load_cert()

    def _get_ca_key(self) -> crypto.PKey:
        """
        Get the CA private key. If the key is passphrase protected, the user is prompted to enter it.
        After the key is retrieved, the CA key object is cleared from the self.ca.

        :return: the CA private key
        """

        if not self.ca:
            self._load_ca()

        self.ca.load_key()
        ca_key = self.ca.key_obj
        self.ca.clear_key()

        return ca_key

    def _get_config(self, config_name: str):
        """
        Get the value from a config parameter.

        The value is casted to the right type depending of the config parameter type in the vault.

        :param config_name: the config name
        :return: the value in int, string or bytes
        """
        c = Config.get_by_id(config_name)
        if not c:
            raise RuntimeError(f'The config name "{config_name}" is unknown')

        if c.type == ConfigType.INT:
            return int(c.value)
        elif c.type == ConfigType.STRING:
            return c.value
        elif c.type == ConfigType.BINARY:
            return c.value.decode('utf8')
        else:
            raise NotImplementedError(f'Config type "{c.type}" not implemented')

    @database.connection_context()
    def select_certificate(self, search_term: str) -> Certificate:
        """
        Select a certificate by specifying a search term or by giving the certificate ID.

        :param search_term: CN part or ID
        :return: the certificate
        """
        if re.match('^[0-9]+$', search_term):
            certificate = Certificate.get_by_id(int(search_term))
        else:
            now = datetime.datetime.now()
            certificates = Certificate.\
                select().where((Certificate.cn.contains(search_term)) & (Certificate.is_renewed == False) &
                               ((Certificate.is_revoked == False) & (Certificate.not_after >= now)))

            if certificates.count() == 1:
                certificate = certificates[0]
            else:
                if certificates.count() == 0:
                    print('No certificate found for this search. Listing all certificates :')
                    certificates = Certificate.select().where((Certificate.is_renewed == False) &
                                                      (Certificate.not_after > datetime.datetime.now()) &
                                                      (Certificate.is_revoked == False))
                else:
                    print(f'{certificates.count()} certificates found :')

                valid_ids = []
                for c in certificates:
                    valid_ids.append(c.id)
                    print_tabulated(str(c))

                print('')
                target_certificate = ask('Please enter the ID of one of these certificates', values=valid_ids)
                certificate = Certificate.get_by_id(int(target_certificate))

        return certificate

    @database.connection_context()
    def _sign_csr(self, csr: crypto.X509Req, overwrite_hashalgo: str = None, overwrite_cn: Optional[str] = None,
                  overwrite_san: Optional[str] = None, overwrite_expire: Optional[str] = None,
                  ask_confirm: bool = False, disable_cert_existence_check: bool = False) -> (crypto.X509, Certificate):
        """
        Sign the Certificate Signing Request with the Certificate Authority.

        The overwrite_hashalgo parameter can be used to overwrite the default hash algorithm.
        The overwrite_cn parameter can be used to overwrite the Common Name in the CSR.
        The overwrite_san parameter can be used to overwrite the Subject Alternative Name in the CSR.
        The overwrite_expire parameter can be used to overwrite the expire date (must be in the future).

        The SAN must always include an entry matching the CN. If the CSR doesn't contain a SAN, this method create it.

        A unique (and non already used) serial number is assigned to the certificate. The certificate is valid for time
        configured in the Config<default_validity_seconds>.

        This method save the certificate in the vault. Note that the private key is not saved because not known. The
        calling function must update the vault with the private key if needed.

        This method returns a tuple :
            * the certificate object
            * the certificate model instance

        :param csr: the Certificate Signing Request
        :param overwrite_hashalgo: the optional hash algorithm. See AVALAIBLE_HASHALGO.
        :param overwrite_cn: the optional Command Name that must be used
        :param overwrite_san: the optional Subject Alternative Name that must be used
        :param overwrite_expire: the optional expired date
        :param ask_confirm: the optional flag to print the CSR and ask the confirmation before continuing
        :param disable_cert_existence_check: disable the certificate existing check
        :return: (the OpenSSL certificate object, the Certificate model instance)
        """
        final_cn = overwrite_cn if overwrite_san else csr.get_subject().CN
        if not regex_domain.match(final_cn):
            raise ValueError(f'CN "{final_cn}" is invalid')

        final_san = validate_or_build_san(final_cn, overwrite_san if overwrite_san else get_csr_san(csr))

        if overwrite_hashalgo and overwrite_hashalgo not in VALID_HASH_ALGOS:
            raise ValueError(f'The hash algorithm "{overwrite_hashalgo}" is invalid')
        hashlago = overwrite_hashalgo if overwrite_hashalgo else self._get_config('default_hash_algo')

        if overwrite_expire:
            try:
                if not isinstance(overwrite_expire, datetime.date):
                    overwrite_expire = datetime.datetime.strptime(overwrite_expire, '%d/%m/%Y').date()

                if overwrite_expire < datetime.date.today():
                    raise ValueError(f'The expire date is not in the future')
            except ValueError:
                raise ValueError(f'The expire date "{overwrite_expire}" is invalid')
        not_after = nb_seconds_to_date(overwrite_expire) if overwrite_expire else \
            self._get_config('default_validity_seconds')

        if not disable_cert_existence_check:
            # Check if a certificate with the same CN already exists
            c = Certificate.select().where((Certificate.cn == final_cn) &
                                           (Certificate.is_renewed == False) &
                                           (Certificate.not_after > datetime.datetime.now()) &
                                           (Certificate.is_revoked == False))

            assert c.count() < 2, 'Critical error : there is more that one actual certificate matching this CN'

            if c.count() == 1:
                raise RuntimeError(f'The following certificate is matching this CN and it\'s not expired, revoked or '
                                   f'renewed : {c[0]}')

        # Generate unique serial number
        is_serial_validated = False
        serial_number_float64 = None
        serial_number_hex = None
        while not is_serial_validated:
            serial_number_float64 = generate_random_serial_number()
            serial_number_hex = serial_number_float64.to_bytes(8, 'big').hex(':')

            # We ensure that the random serial number is not already used
            is_serial_validated = Certificate.select().where(Certificate.serial == serial_number_hex).count() == 0

        # Load CA
        self._load_ca()

        # Generate the final certificate
        cert = crypto.X509()
        cert.set_version(2)

        cert.set_serial_number(serial_number_float64)
        cert.get_subject().CN = final_cn

        cert.gmtime_adj_notBefore(0)
        cert.gmtime_adj_notAfter(not_after)

        cert.set_issuer(self.ca.cert_obj.get_subject())
        cert.set_pubkey(csr.get_pubkey())

        define_san(cert, final_san)

        if ask_confirm:
            print('You are about to sign the following certificate :')
            print_tabulated(get_cert_text(cert))

            if not confirm('Do you want to sign it'):
                exit(0)

        cert.sign(self._get_ca_key(), hashlago)

        # Persist
        c = Certificate.create(
            cn=final_cn,
            san=final_san,
            created_timestamp=datetime.datetime.now().replace(microsecond=0),
            not_after=datetime.datetime.strptime(cert.get_notAfter().decode('utf8'), '%Y%m%d%H%M%SZ'),
            serial=serial_number_hex,
            cert=crypto.dump_certificate(FILETYPE_PEM, cert).decode('utf8'),
            key=None,
            is_revoked=False,
            is_renewed=False,
        )

        return cert, c

    @database.connection_context()
    def _generate_signed_cert(self, cn: str, san: Optional[str] = None, overwrite_keysize: int = None,
                              overwrite_hashalgo: str = None, overwrite_expire: Optional[str] = None,
                              ask_confirm: bool = False) -> (crypto.X509, Certificate):
        """
        Generate a Certificate and signed it with the Certificate Authority.

        The cn parameter is the Common Name. See Utils.regex_domain for valid values.
        The optional san parameter is the Subject Alternative Name. If not defined, the SAN will be generate by this
        method.

        The overwrite_keysize parameter can be used to overwrite default key size. See AVALAIBLE_KEYSIZES.
        The overwrite_hashalgo parameter can be used to overwrite the default hash algorithm.
        The overwrite_expire parameter can be used to overwrite the number of days of validity.

        The SAN must always include an entry matching the CN. If the CSR doesn't contain a SAN, this method create it.

        A unique (and non already used) serial number is assigned to the certificate. The certificate is valid for time
        configured in the Config<default_validity_seconds>.

        This method save the certificate in the vault. Note that the private key is not saved because not known. The
        calling function must update the vault with the private key if needed.

        This method returns a tuple :
            * the certificate object
            * the certificate model instance

        :param cn: the Common Name
        :param san: the optional Subject Alternative Name
        :param overwrite_keysize: the optional key size. See AVALAIBLE_KEYSIZES for valid values
        :param overwrite_hashalgo: the optional hash algorithm. See AVALABILE_HASHALGOS
        :param overwrite_expire: the optional number of days of validity that must be used
        :param ask_confirm: the optional flag to print the CSR and ask the confirmation before continuing
        :return: (the OpenSSL certificate object, the Certificate model instance)
        """
        if not regex_domain.match(cn):
            raise ValueError(f'CN "{cn}" is invalid')

        san = validate_or_build_san(cn, san)

        if overwrite_keysize and overwrite_keysize not in VALID_KEY_SIZES:
            raise ValueError(f'The key size "{overwrite_keysize}" is invalid')
        overwrite_keysize = overwrite_keysize if overwrite_keysize else self._get_config('default_key_size')

        if overwrite_hashalgo and overwrite_hashalgo not in VALID_HASH_ALGOS:
            raise ValueError(f'The hash algorithm "{overwrite_hashalgo}" is invalid')
        hashalgo = overwrite_hashalgo if overwrite_hashalgo else self._get_config('default_hash_algo')

        # Generate key
        key = crypto.PKey()
        key.generate_key(crypto.TYPE_RSA, overwrite_keysize)

        # Generate CSR
        req = crypto.X509Req()
        req.get_subject().CN = cn
        req.set_pubkey(key)
        req.sign(key, hashalgo)

        # Define the SAN
        define_san(req, san)

        # Sign the CSR
        signed_cert, vault_certificate = self._sign_csr(req,
                                                        overwrite_hashalgo=hashalgo,
                                                        overwrite_expire=overwrite_expire,
                                                        ask_confirm=ask_confirm)

        # Save the private key into the vault
        vault_certificate.key = crypto.dump_privatekey(FILETYPE_PEM, key).decode('utf8')
        vault_certificate.save()

        return signed_cert, vault_certificate



import os.path
from enum import IntEnum
import datetime
import ipaddress
import secrets
import sys
from typing import Optional
import re
import getpass

from OpenSSL import crypto
from OpenSSL.SSL import FILETYPE_PEM

"""
The RegEx to match a domain. Examples :
  - wiki
  - wiki.internal-domain.lan
  - *.sub.internal-domain.lan
"""
regex_domain = re.compile(r'^(\*\.)?([a-zA-Z0-9_-]+\.?)+$')

"""
The RegEx to match a Subject Alternate Name.

  - DNS: host, DNS: host2.domain.lan, IP: 10.0.0.1, IP: 2001:0db8:85a3:0000:0000:8a2e:0370:7334
"""
regex_san = re.compile(r'^(('
                       r'(DNS:[ ]?((\*\.)?([a-zA-Z0-9_-]+\.?))+)'  # DNS, see regex_domain
                       r'|'
                       r'(IP:[ ]?('  # IPv4
                       r'((?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))'
                       r'|'  # or IPv6
                       r'((?:[a-fA-F0-9]{1,4}:){7}[a-fA-F0-9]{1,4})'
                       r')))'
                       r'(,?[ ]?))+'  # Followed by "," or ", "
                       r'(?<![ ,])$')  # But the string cannot be ending with "," or ", "

"""
The RegEx for the Common Name : host or IPv4 or IPv6
"""
regex_cn = re.compile(r'^(((\*\.)?([a-zA-Z0-9_-]+\.?)+)|'
                      r'((?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))|'
                      r'((?:[a-fA-F0-9]{1,4}:){7}[a-fA-F0-9]{1,4}))$')


def ask(prompt: str, values: Optional[list] = None, disable_show_values: bool = False, can_be_empty: bool = False,
        disable_lowering: bool = False) -> Optional[str]:
    """
    Ask the user for an input.

    The prompt is written to stdout and the value is read from stdin.

    The prompt will be like that :
        "<prompt message> : "
        "<prompt message> (<value 1>, <value 2>, <value 3>) : "

    The optional disable_show_values parameter disable showing values in the prompt.

    The optional values parameter is the fixed list of the authorized values. If defined, the prompt is asking until the
    user is entering one of these values.

    The optional can_be_empty parameter allows the user to not enter a value but send back a empty value.

    By default, the input string is lower(). If you doesn't want that, you can use the disable_lowering parameter.

    :param prompt: the prompt message
    :param values: the list of valid values
    :param disable_show_values: disable showing values
    :param can_be_empty: indicate if the response can be empty
    :param disable_lowering: disable the lower() of the response
    :return: the user response
    """
    if values:
        values = [str(v).lower() for v in values]

        if not disable_show_values:
            prompt += f' ({", ".join(values)})'

    prompt += ' : '

    while True:
        print(prompt, end='')
        answer = input().strip()
        answer_lower = answer.lower()

        if values:
            if answer_lower in values or (can_be_empty and len(answer) == 0):
                break
        else:
            if (len(answer) == 0 and can_be_empty) or len(answer) > 0:
                break

    if len(answer) == 0:
        return None

    if disable_lowering:
        return answer
    else:
        return answer_lower


class PEMFormatType(IntEnum):
    """
    The PEM type
    """
    CERT = 1
    KEY = 2
    CSR = 3


def ask_pem_format(prompt: str, pem_type: PEMFormatType) -> Optional[str]:
    """
    Ask the user for data in the PEM format.

    The targeted data is providing through the pem_type parameter.

    This function accepts multi-line PEM. The first line must be the ASCII armored definition "----- BEGIN...". This
    function will stop reading from stdin when the the "----- END..." is reached.

    The PEM data is not checked.

    :param prompt: the prompt message
    :param pem_type: the PEM format type
    :return: the entered PEM data
    """
    answer = ''

    prompt += ' :'
    print(prompt)

    if pem_type == PEMFormatType.CERT:
        re_start_line = re.compile('-----BEGIN CERTIFICATE-----')
        re_end_line = re.compile('-----END CERTIFICATE-----')
    elif pem_type == PEMFormatType.KEY:
        re_start_line = re.compile('-----BEGIN (RSA )?PRIVATE KEY-----')
        re_end_line = re.compile('-----END (RSA )?PRIVATE KEY-----')
    elif pem_type.CSR:
        re_start_line = re.compile('-----BEGIN CERTIFICATE REQUEST-----')
        re_end_line = re.compile('-----END CERTIFICATE REQUEST-----')
    else:
        raise RuntimeError(f'PEM type {pem_type} not supported')

    end_line_detected = False
    first = True
    while 1:
        try:
            line = sys.stdin.readline()
        except KeyboardInterrupt:
            break

        if first:
            first = False
            if not re_start_line.match(line.rstrip()):
                raise ValueError('No or bad start line')

        answer += line

        if re_end_line.match(line.rstrip()):
            end_line_detected = True
            break

    if not end_line_detected:
        raise ValueError('No end line detected')

    return answer


def ask_date_in_future() -> datetime.date:
    """
    Ask the user for a date in the future.

    If an error is detected, the error is written on stdout and the same question is prompted again.

    :return: the date
    """
    error = True
    d = ''

    while error:
        d = ask('Enter a date in the future (DD/MM/YYYY)')

        try:
            d = datetime.datetime.strptime(d, '%d/%m/%Y').date()

            if d < datetime.date.today():
                print(f'\tThe date must be in the future')
            else:
                error = False
        except ValueError:
            print(f'\t{d} is invalid')

    return d


def ask_cn() -> str:
    """
    Ask the user for Common Name.

    The user can enter 1 host or IP.

    If an error is detected, the error is written on strerr and the same question is prompted again.

    :return: the CN
    """
    error = True
    cn = ''

    while error:
        cn = ask('Common Name')

        if not regex_cn.match(cn):
            print(f'\t{cn} is invalid')
        else:
            error = False

    return cn


def ask_hosts() -> [str]:
    """
    Ask the user for hosts.

    The user can enter 0, 1 or n hosts separated by a comma. The host value can be :
        * the hostname : wiki
        * the FQDN : wiki.internal-domain.lan
        * the * wildcard : *.sub.internal-domain.lan

    If an error is detected, the error is written on strerr and the same question is prompted again.

    :return: the hosts
    """
    error = True
    domains = set()

    while error:
        input_domains = ask('Enter alternative names separated by a comma (eg: srv.internal.lan)', can_be_empty=True)
        domains.clear()

        if not input_domains or len(input_domains) == 0:
            break

        error = False
        for d in input_domains.split(','):
            d = d.strip()

            if regex_domain.match(d):
                try:
                    ipaddress.ip_address(d)
                    error = True
                    print(f'\t{d} is an IP address')
                except ValueError:
                    domains.add(d)
            else:
                error = True
                print(f'\t{d} is invalid')

    return list(sorted(domains))


def ask_ips() -> [str]:
    """
    Ask the user for IP addresses.

    The user can enter 0, 1 or n IP addresses separated by a comma. IPv4 and IPv6 are supported.

    If an error is detected, the error is written on strerr and the same question is prompted again.

    :return: the IPs
    """
    error = True
    ips = set()

    while error:
        input_ips = ask('Enter IPs separated by a comma (eg: 10.0.0.1, 10.254.0.1)', can_be_empty=True)
        ips.clear()

        if not input_ips or len(input_ips) == 0:
            break

        error = False
        for ip in input_ips.split(','):
            ip = ip.strip()

            try:
                ipaddress.ip_address(ip)
                ips.add(ip)
            except ValueError:
                error = True
                print(f'\t{ip} is invalid')

    return list(sorted(ips))


def ask_cn_and_san() -> (str, str):
    """
    Ask the user for the CN, the alternative hosts and alternative IPS.

    :return: the CN and the build SAN
    """
    cn = ask_cn()
    alternate_hosts = ask_hosts()
    alternate_ips = ask_ips()

    try:  # CN is IP
        ipaddress.ip_address(cn)
        alternate_ips.append(cn)
    except ValueError:
        alternate_hosts.append(cn)

    san = [f'DNS:{h}' for h in alternate_hosts]
    san += [f'IP:{i}' for i in alternate_ips]
    san = ', '.join(san)

    return cn, san


def confirm(prompt: str) -> bool:
    """
    Ask user to enter Y or N (case-insensitive) to confirm an action.

    :param prompt: the prompt message
    :return: True if the answer is Y.
    """
    return ask(f'{prompt} ? [y/n]', ['y', 'n'], disable_show_values=True) == 'y'


def ask_password(prompt: str) -> str:
    """
    Ask (securely) the user to enter a password.

    :param prompt: the prompt message
    :return: the password encoded
    """
    prompt += ' : '

    return getpass.getpass(prompt)


def ask_private_key_passphrase(*args) -> str:
    """
    Ask the user for the private key passphrase.

    :param args:
    :return: the passsphrase
    """
    return ask_password('Enter the private key passphrase')


def generate_random_serial_number() -> int:
    """
    Generate (securely) a random serial number for a certificate [0, 2^64].
    
    :return: the serial number
    """
    return secrets.randbelow(2 ** 64)


def get_cert_san(cert: crypto.X509) -> Optional[str]:
    """
    Get the Subject Alternative Name from a certificate object.

    :param cert: the certificate
    :return: the SAN if defined
    """
    for i in range(0, cert.get_extension_count()):
        if cert.get_extension(i).get_short_name() == b'subjectAltName':
            return str(cert.get_extension(i)).replace('IP Address', 'IP')

    return None


def get_csr_san(csr: crypto.X509Req) -> Optional[str]:
    """
    Get the Subject Alternative Name from a certificate signing request object.

    :param csr: the certificate signing request
    :return: the SAN if defined
    """
    for ext in csr.get_extensions():
        if ext.get_short_name() == b'subjectAltName':
            return str(ext).replace('IP Address', 'IP')

    return None


def get_csr_cn(csr: crypto.X509Req) -> str:
    """
    Get the Common Name from a certificate signing request object.

    :param csr: the certificate signing request
    :return: the CN
    """
    return csr.get_subject().CN


def define_san(cert_or_csr: crypto.X509 or crypto.X509Req, san: str):
    """
    Define the Subject Alternative Name for a certificate or a certificate signing request.

    :param cert_or_csr: a certificate or a certificate signing request.
    :param san: the Subject Alternative Name
    :return:
    """
    if not regex_san.match(san):
        raise ValueError(f'The SAN "{san}" is invalid')

    san_already_defined = False
    if isinstance(cert_or_csr, crypto.X509):
        if get_cert_san(cert_or_csr):
            san_already_defined = True
    else:
        if get_csr_san(cert_or_csr):
            san_already_defined = True

    if san_already_defined:
        raise RuntimeError('Cannot define SAN because the extension is already defined')

    cert_or_csr.add_extensions([
        crypto.X509Extension(b'subjectAltName',
                             False,
                             san.encode('utf8'))])

    return cert_or_csr


def validate_or_build_san(cn: str, san: Optional[str]) -> str:
    """
    Validate the Subject Alternative Name depending on the Common Name. If the SAN is empty, this function creates one
    depending of the Common Name.

    :param cn: the Common Name
    :param san: the optional Subject Alternative Name
    :return: the final Subject Alternative Name
    """
    # Checking the the CN is an IP address or a hostname
    try:
        ipaddress.ip_address(cn)
        cn_is_ip = True
    except ValueError:
        cn_is_ip = False

    if san:
        values = [val.strip().replace(' ', '') for val in san.split(',')]
    else:
        values = []

    if cn_is_ip:
        cn_san = f'IP:{cn}'
    else:
        cn_san = f'DNS:{cn}'

    if cn_san not in values:
        values.append(cn_san)

    san = ', '.join(values)

    if not regex_san.match(san):
        raise ValueError(f'The SAN "{san}" is invalid')

    return san


def get_cert_text(cert: crypto.X509) -> str:
    """
    Get the text version of a certificate object.

    :param cert: the certificate
    :return: the text version
    """
    return crypto.dump_certificate(2 ** 16 - 1, cert).decode('utf8')


def get_csr_text(csr: crypto.X509Req) -> str:
    """
    Get the text version of a certificate signing request object.

    :param csr: the certificate signing request
    :return: the text version
    """
    return crypto.dump_certificate_request(2 ** 16 - 1, csr).decode('utf8')


def get_private_key_text(key: crypto.PKey) -> str:
    """
    Get the text version of a private key object.

    :param key: the private key
    :return: the text version
    """
    return crypto.dump_privatekey(2 ** 16 - 1, key).decode('utf8')


def nb_seconds_to_date(expire_date: datetime.date) -> int:
    """
    Compute the number of seconds between now and the expired date. Note that the date is given without the time part
    so we use the current hours/minutes/seconds.

    :param expire_date: the expire date
    :return: the number of seconds
    """
    if expire_date < datetime.date.today():
        raise ValueError('The expire date is not in the future')

    expire_dt = datetime.datetime.now().replace(year=expire_date.year, month=expire_date.month, day=expire_date.day)
    return int((expire_dt - datetime.datetime.now()).total_seconds())


def print_tabulated(text: str) -> None:
    """
    Print the text with one tabulation.

    :param text: the text to print
    """

    text = '\t' + text.replace('\n', '\n\t')
    print(text)


def write_cert_pem(cert: crypto.X509, filepath: str):
    """
    Write the certificate in PEM format into the provided filepath.

    If the file already exists, a RuntimeError is raised.

    :param cert: the certificate
    :param filepath: the target filepath
    """
    if os.path.exists(filepath):
        raise RuntimeError(f'The filepath "{filepath}" already exists')

    with open(filepath, 'w') as f:
        f.write(crypto.dump_certificate(FILETYPE_PEM, cert).decode('utf8'))


def write_private_key_pem(key: crypto.PKey, filepath: str):
    """
    Write the private key in PEM format into the provided filepath.

    If the file already exists, a RuntimeError is raised.

    :param key: the private key
    :param filepath: the target filepath
    """
    if os.path.exists(filepath):
        raise RuntimeError(f'The filepath "{filepath}" already exists')

    with open(filepath, 'w') as f:
        f.write(crypto.dump_privatekey(FILETYPE_PEM, key).decode('utf8'))


def write_p12(filepath: str, passphrase: str, cert: crypto.X509, key: Optional[crypto.PKey] = None):
    """
    Write the certificate (and the private key if provided) in PKCS12 format.

    If the file already exists, a RuntimeError is raised.

    :param filepath: the target filepath
    :param passphrase: the password
    :param cert: the certificate
    :param key: the private key
    """
    if os.path.exists(filepath):
        raise RuntimeError(f'The filepath "{filepath}" already exists')

    if len(passphrase) == 0:
        raise ValueError('The provided passphrase is empty')

    crypto.PKCS12()
    pfx = crypto.PKCS12()
    pfx.set_certificate(cert)

    if key:
        pfx.set_privatekey(key)

    pfx_data = pfx.export(passphrase)
    with open(filepath, 'wb') as f:
        f.write(pfx_data)



import argparse
import os
import shutil
import sys

from .camanager import CAManager
from .models import DATABASE_FILENAME, DATABASE_FILENAME_BACKUP
from .utils import confirm

def main():
    parser = argparse.ArgumentParser(formatter_class=argparse.RawDescriptionHelpFormatter, prog='camanager')

    subparsers = parser.add_subparsers(dest='actions', help='The action')
    setup_subparsers = subparsers.add_parser('setup', help='Initialize the vault')
    backup_subparsers = subparsers.add_parser('backup', help='Backup the vault')
    restore_subparsers = subparsers.add_parser('restore', help='Restore the vault from the backup')
    newcert_subparsers = subparsers.add_parser('newcert', help='Create a new certificate')
    sign_subparsers = subparsers.add_parser('sign', help='Sign a certificate')
    renew_subparsers = subparsers.add_parser('renew', help='Renew a certificate')
    list_subparsers = subparsers.add_parser('list', help='List certificates')
    export_subparsers = subparsers.add_parser('export', help='Export a certificate')
    subparsers.required = True

    newcert_subparsers.add_argument('--cn', help=argparse.SUPPRESS)
    newcert_subparsers.add_argument('--san', help=argparse.SUPPRESS)
    newcert_subparsers.add_argument('--keysize', help=argparse.SUPPRESS)
    newcert_subparsers.add_argument('--hash', help=argparse.SUPPRESS)

    sign_subparsers.add_argument('--cn', help=argparse.SUPPRESS)
    sign_subparsers.add_argument('--san', help=argparse.SUPPRESS)
    sign_subparsers.add_argument('csr_file', nargs='?', help='The CSF filepath')

    renew_subparsers.add_argument('certificate', help='The certificate # or Common Name')

    list_group = list_subparsers.add_mutually_exclusive_group()
    list_group.add_argument('--all_certificates', action='store_true', help='Also the revoked/renewed/expired '
                                                                            'certificates')
    list_group.add_argument('--soon-expired', action='store_true', help='Only the soon expired (> 1 month)')

    export_group = export_subparsers.add_mutually_exclusive_group(required=True)
    export_group.add_argument('--pem', action='store_true', help='PEM output')
    export_group.add_argument('--p12', action='store_true', help='P12 output')
    export_subparsers.add_argument('--out', nargs='?', help='The output file')
    export_subparsers.add_argument('certificate', help='The certificate # or Common Name')

    args = parser.parse_args()

    cam = CAManager()

    if args.actions == 'setup':
        try:
            os.remove(DATABASE_FILENAME)
        except FileNotFoundError:
            pass

        if os.path.isfile(DATABASE_FILENAME):
            msg = f'Error : the database "{DATABASE_FILENAME}" already exists.\n\n' \
                  f'You can remove this file BUT YOU ARE GOING TO LOST EVERYTHING RELATED TO THE CURRENT CERTIFICATE ' \
                  f'AUTHORITY. Be sure of the consequence of your action.'

            sys.stderr.write(f'The database "{DATABASE_FILENAME}" already exists. ')
            exit(-1)

        cam.create_vault()

    elif args.actions == 'backup':
        if not os.path.isfile(DATABASE_FILENAME):
            msg = f'Error : the database "{DATABASE_FILENAME}" not found.' \

            print(msg)
            exit(-1)

        if os.path.isfile(DATABASE_FILENAME_BACKUP):
            msg = f'Error : the database backup "{DATABASE_FILENAME_BACKUP}" already exists.\n\n' \
                  f'You can remove this file BUT YOU ARE GOING TO LOST EVERYTHING RELATED TO THE BACKUP.'

            print(msg)

            if not confirm('Do you to ERASE the previous backup'):
                print('No backup was done')
                exit(0)
            else:
                os.remove(DATABASE_FILENAME_BACKUP)

        shutil.copy(DATABASE_FILENAME, DATABASE_FILENAME_BACKUP)
        print('Backup done successfully')

    elif args.actions == 'restore':
        if not os.path.isfile(DATABASE_FILENAME_BACKUP):
            sys.stderr.write(f'No backup found')
            exit(-1)

        if os.path.isfile(DATABASE_FILENAME):
            msg = f'Warning : YOU ARE GOING TO LOST EVERYTHING RELATED TO THE CURRENT CERTIFICATE AUTHORITY. ' \
                  f'Be sure of the consequence of your action.\n'

            print(msg)

            if not confirm('Do you you want to restore the backup'):
                print('The backup has NOT been restored')
                exit(0)
            else:
                os.remove(DATABASE_FILENAME)

        shutil.copy(DATABASE_FILENAME_BACKUP, DATABASE_FILENAME)

        print('Backup restored successfully')

        if not confirm('Do you want to keep the backup'):
            os.remove(DATABASE_FILENAME_BACKUP)
            print('Backup file deleted successfully')

    elif args.actions == 'newcert':
        cam.load()
        cam.generate_new_cert()

    elif args.actions == 'sign':
        cam.load()
        cam.sign_csr(args.csr_file, args.cn, args.san)

    elif args.actions == 'renew':
        cam.load()
        cam.renew(args.certificate)
        # cam.sign_csr(args.csr_file)

    elif args.actions == 'list':
        cam.load()
        cam.list(args.all_certificates, args.soon_expired)

    elif args.actions == 'export':
        cam.load()
        cam.export(args.certificate, "p12" if args.p12 else "pem", args.out)


if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        print('\n\n*** Interrupted by user ***')



import datetime
import enum

import peewee
from playhouse.sqlcipher_ext import SqlCipherDatabase
from OpenSSL import crypto
from OpenSSL.SSL import FILETYPE_PEM

from .utils import ask_private_key_passphrase

"""
The database filename
"""
DATABASE_FILENAME = 'ca.vault'  # ca.vault

"""
The database filename backup
"""
DATABASE_FILENAME_BACKUP = f'{DATABASE_FILENAME}.bak'

"""
The database object
"""
database = SqlCipherDatabase(None)


class BaseModel(peewee.Model):
    """
    Base model
    """
    class Meta:
        database = database


class BaseCertKeyModel(BaseModel):
    """
    Base certificate model which may have a key.

    The certificate must be stored into the "cert" attribute in the PEM format.
    The optional private key must be stored into the "key" attribute in the PEM format.

    The OpenSSL.crypto.X509 and OpenSSL.crypto.Pkey can be loaded as self.cert_obj and self.key_obj. If the key is
    passphrase protected, the user is asking for this passphrase.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.cert_obj = None
        self.key_obj = None

    def load_cert(self) -> None:
        """
        Load the certificate as OpenSSL.crypto.X509 object into the attribute "cert_obj".
        """
        self.cert_obj = crypto.load_certificate(FILETYPE_PEM, self.cert.encode('utf8'))

    def load_key(self) -> None:
        """
        Load the private key as OpenSSL.crypto.Pkey object into the attribute "key_obj".

        If the key is passphrase protected, the user is asking for this passphrase (prompt to stdout, read from stdin).

        If no key is available (self.key is None), a RuntimeError is raised.

        :except: RunTime
        """
        if not self.key:
            raise RuntimeError('No key available')

        self.key_obj = crypto.load_privatekey(FILETYPE_PEM, self.key.encode('utf8'),
                                              passphrase=ask_private_key_passphrase)

    def clear_key(self) -> None:
        """
        Clear the attribute "key_obj".
        """
        self.key_obj = None


class CA(BaseCertKeyModel):
    """
    A Certificate Authority model.

    The certificate and the key are stored in the PEM format.
    """
    id = peewee.IntegerField(primary_key=True)
    name = peewee.CharField(unique=True)
    cert = peewee.TextField()
    key = peewee.TextField()


class CertificateRevokedReason(enum.Enum):
    """
    The certificate revoked reason
    """
    UNSPECIFIED = 'unspecified'
    KEY_COMPRISE = 'keyCompromise'
    CESSATION_OF_OPERATION = 'cessationOfOperation'


class Certificate(BaseCertKeyModel):
    """
    A certificate model.

    The created_timestamp and not_after are precise to the second.
    The serial is provided in the official X509 output, so in hex format using ":" separator in the big endian
    representation of the 64 bits integer.
    The certificate and the key (optional) are stored in the PEM format.
    """
    id = peewee.IntegerField(primary_key=True)
    cn = peewee.CharField()
    san = peewee.CharField(null=True)
    created_timestamp = peewee.DateTimeField()
    not_after = peewee.DateTimeField()
    serial = peewee.CharField()
    cert = peewee.TextField()
    key = peewee.TextField(null=True)
    is_revoked = peewee.BooleanField()
    revoked_timestamp = peewee.DateTimeField(null=True)
    revoked_reason = peewee.BlobField(null=True)
    revoked_comment = peewee.TextField(null=True)
    is_renewed = peewee.BooleanField()
    renewed_cert_id = peewee.IntegerField(null=True)

    def __str__(self):
        now = datetime.datetime.now()

        s = f'#{self.id}. CN:{self.cn}'\

        if self.san:
            s += f' / {self.san}'

        s += ' ('

        if self.is_renewed:
            s += 'renewed'
        elif self.is_revoked:
            s += 'revoked'

            if self.revoked_comment:
                s += f' [reason: {self.revoked_comment}]'
        elif self.not_after <= now:
            s += 'expired'
        else:
            s += f'expire on {self.not_after.strftime("%d/%m/%Y %H:%M:%S")}'

        s += f', serial : {self.serial}'

        if not self.key:
            s += ', no private key'

        s += ')'

        return s


class CRL(BaseModel):
    """
    A Certificate Revoked List model

    The crl_timestamp is precise to the second.
    """
    id = peewee.IntegerField(primary_key=True)
    crl_version = peewee.IntegerField()
    crl_timestamp = peewee.DateTimeField()
    crl = peewee.TextField()


class ConfigType(enum.IntEnum):
    """
    The configuration value type
    """
    INT = 1
    STRING = 2
    BINARY = 3


class Config(BaseModel):
    """
    A Configuration model
    """
    name = peewee.CharField(primary_key=True)
    type = peewee.IntegerField()
    value = peewee.CharField()
